#ifndef __tag_h_
#define __tag_h_

#include <sys/types.h>
#include <mywchar.h>

#include "dtd.h"

enum {
	TAG_OPEN, TAG_CLOSE
};

typedef enum {
	TAG_none,
	TAG_comment,
	TAG_normal,
	TAG_string,
} TagType;

typedef enum {
	TAG_unknown,

	/* text data */
	TAG_pcdata,

	/* fontstyle */
	TAG_tt,
	TAG_b,
	TAG_i,
	TAG_big,
	TAG_small,

	/* phrase */
	TAG_em,
	TAG_strong,
	TAG_dfn,
	TAG_code,
	TAG_samp,
	TAG_kbd,
	TAG_var, 
	TAG_cite, 
	TAG_abbr,
	TAG_acronym,

	/* special */
	TAG_sub,
	TAG_sup,
	TAG_span,
	TAG_bdo,

	TAG_br,

	/* body */
	TAG_body,
	TAG_address,
	TAG_div,

	/* anchor */
	TAG_a, 

	/* image map */
	TAG_map,
	TAG_area,

	/* link */
	TAG_link,

	/* images */
	TAG_img,

	/* object */
	TAG_object,
	TAG_param,

	/* horizontak rule */
	TAG_hr,

	/* paragraph */
	TAG_p,

	/* heading */
	TAG_h1, TAG_h2, TAG_h3, TAG_h4, TAG_h5, TAG_h6,

	TAG_pre,
	TAG_q, TAG_blockquote,
	TAG_ins, TAG_del,

	/* lists */
	TAG_dl, TAG_dt, TAG_dd,
	TAG_ol, TAG_ul, TAG_li,

	/* form */
	TAG_form, TAG_label,
	TAG_input, TAG_select, 
	TAG_optgroup, TAG_option,
	TAG_textarea,
	TAG_fieldset,
	TAG_legend,
	TAG_button,

	/* table */
	TAG_table, TAG_caption,
	TAG_thead, TAG_tfoot, TAG_tbody,
	TAG_colgroup, TAG_col,
	TAG_tr, TAG_th, TAG_td,

	/* document head */
	TAG_head,
	TAG_title, 
	TAG_base,
	TAG_meta,
	TAG_style,
	TAG_script,
	TAG_noscript,

	/* html */
	TAG_html,

	/* parse base */
	TAG__start_


} TAGS;

typedef enum {
	OPT_unknown,
	OPT_background, OPT_bgcolor, OPT_text, OPT_alink, OPT_link, OPT_vlink,
	OPT_name, OPT_type, OPT_start, OPT_size, OPT_color,
	OPT_noshade,
	OPT_width, OPT_height, OPT_compact,
	OPT_href,
	OPT_src, OPT_lowsrc, OPT_alt, OPT_border,
	OPT_hspace, OPT_vspace,
	OPT_align, OPT_valign,
	OPT_method, OPT_action,
	OPT_value, OPT_maxlength, OPT_checked, OPT_multiple, OPT_selected,
	OPT_rows, OPT_cols, OPT_cellpadding, OPT_cellspacing,
	OPT_nowrap, OPT_rowspan, OPT_colspan,
	OPT_left, OPT_center, OPT_right, 
	OPT_top, OPT_bottom, OPT_middle, OPT_absmiddle,
	OPT_clear, OPT_all, OPT_none,
	OPT_disc, OPT_square, OPT_circle,
	OPT_id,
	OPT_class
} OPTS;

typedef enum {
	TAG_NODISP,
	TAG_INLINE,
	TAG_BLOCK,
	TAG_PSEUDO
} TagFormatType;

typedef struct TagInfo {
	struct TagInfo *next;		
	DTDObjList dtd;				/* DTD ʸˡ */
	char *name;					/* ̾ 		*/
	TAGS id;					/*  ID  		*/
	void (*parse_func)();		/* ѡѴؿ */
	int ommit;					/* ά     */
	int block;					/* ֥åʥ   */
} TagInfo;

/* tag.c */
void tag_init(void);
TagInfo* GetTagInfo(char *s);
extern TagInfo taginfo_pcdata;

typedef struct OptInfo {
	char *name;
	OPTS id;
	struct OptInfo *next;
} OptInfo;
OptInfo *GetOptInfo(char *s);

typedef struct EntList {
    char *name;
    u_short	code;
    struct EntList *next;
} EntList;
EntList *GetEntity(char *s);

typedef struct TagOpt {
	struct TagOpt *next;
	OptInfo *oi;
	char *val;
} TagOpt, *TagOptList;

typedef struct Tag {
	struct Tag *next;
	struct Tag *prev;
	int mode;
	TagType type;
	TagInfo *ti;
	TagOptList opt;
	void *data;
} Tag, *TagList;

Tag* tag_new(TagType);
void tag_delete(Tag *p);
char* tag_get_opt_value(Tag *tag, OPTS id);


typedef struct TagWork {
	int mode;

	int fd;				/* ɤߤǥץ */
	int eof;			/* ɤ߹ߴλե饰		*/

	char *data;			/* multibyte text */
	int pre;			/* pre-formatted mode */
	int tagend;			/* ľե饰 */			

	char *p;			/* current point  */
	char *p2;
	mbstate_t state;	/* state mb->wc   */
	
	TagList tags;		/* tagѡλߥꥹ */
	Tag *cur;			/* html ѡݥ */

	wchar_t (*GetChar)(struct TagWork *);
	int (*main)(struct TagWork *);
} TagWork;

TagWork *tagwork_open(int fd);
void tagwork_delete(TagWork *tw);
void tagwork_reset(TagWork *tw);
Tag *tagwork_gettag(TagWork *w);

#endif
