#include <string.h>
#include <ctype.h>

#include "tag.h"
#include "parse.h"
#include "stylesheet.h"

#define HASHSIZE 161
static TagInfo *hashtab[HASHSIZE];

#define TAGINFO(tag)  NULL,NULL,#tag,TAG_ ## tag
#define TAGINFO1(tag) NULL,NULL,#tag,TAG_ ## tag,frame_parse_ ## tag
#define TAGINFO2(tag) NULL,NULL,#tag,TAG_ ## tag,NULL

TagInfo taginfo_pcdata = { TAGINFO2(pcdata), 0, TAG_INLINE };

static TagInfo taginfo_table[] = {

	/* ================ Text Markup ================ */

	/* fontstyle */
	{ TAGINFO2(tt), 0, TAG_INLINE },
	{ TAGINFO(b), frame_parse_fontstyle, 0, TAG_INLINE },
	{ TAGINFO(i), frame_parse_fontstyle, 0, TAG_INLINE },
	{ TAGINFO2(big), 0, TAG_INLINE },
	{ TAGINFO2(small), 0, TAG_INLINE },

	/* phrase */
	{ TAGINFO2(em), 0, TAG_INLINE },
	{ TAGINFO2(strong), 0, TAG_INLINE },
	{ TAGINFO2(dfn), 0, TAG_INLINE },
	{ TAGINFO2(code), 0, TAG_INLINE },
	{ TAGINFO2(samp), 0, TAG_INLINE },
	{ TAGINFO2(kbd), 0, TAG_INLINE },
	{ TAGINFO2(var), 0, TAG_INLINE },
	{ TAGINFO2(cite), 0, TAG_INLINE },
	{ TAGINFO2(abbr), 0, TAG_INLINE },
	{ TAGINFO2(acronym), 0, TAG_INLINE },

	/* special */
	{ TAGINFO2(sub), 0, TAG_INLINE },
	{ TAGINFO2(sup), 0, TAG_INLINE },
	{ TAGINFO2(span), 0, TAG_INLINE },
	{ TAGINFO2(bdo), 0, TAG_INLINE },

	{ TAGINFO1(br), 0, TAG_INLINE },

	/* ===========  HTML content models ============== */

	/* ==== document body ==== */
	{ TAGINFO1(body), 3, TAG_BLOCK },
	{ TAGINFO2(address), 0, TAG_BLOCK },
	{ TAGINFO2(div), 0, TAG_BLOCK },

	/* ==== anchor element ==== */

	{ TAGINFO1(a), 0, TAG_INLINE },

	/* ==== CTient side image maps ==== */

	{ TAGINFO2(map), 0, TAG_INLINE },
	{ TAGINFO2(area), 2, TAG_INLINE },

	/* ==== link element ==== */
	{ TAGINFO2(link), 2, TAG_NODISP },

	/* ==== images ==== */
	{ TAGINFO1(img),  2, TAG_NODISP },

	/* ==== object ==== */
	{ TAGINFO2(object),  2, TAG_NODISP },
	{ TAGINFO2(param),  2, TAG_NODISP },

	/* ==== hirizopntal rule ==== */
	{ TAGINFO1(hr), 2, TAG_BLOCK },

	/* ==== paragraph ==== */
	{ TAGINFO1(p),   2, TAG_BLOCK },
	
	/* ==== headings ==== */
	{ TAGINFO(h1), frame_parse_heading, 0, TAG_BLOCK },
	{ TAGINFO(h2), frame_parse_heading, 0, TAG_BLOCK },
	{ TAGINFO(h3), frame_parse_heading, 0, TAG_BLOCK },
	{ TAGINFO(h4), frame_parse_heading, 0, TAG_BLOCK },
	{ TAGINFO(h5), frame_parse_heading, 0, TAG_BLOCK },
	{ TAGINFO(h6), frame_parse_heading, 0, TAG_BLOCK },

	/* ==== preformatted text ==== */
	{ TAGINFO1(pre), 0, TAG_BLOCK },

	/* ==== inline quotes ==== */
	{ TAGINFO2(q), 0, TAG_INLINE },

	/* ==== block-like quotes ==== */
	{ TAGINFO2(blockquote), 0, TAG_BLOCK },

	/* === inserter/deleted text ==== */
	{ TAGINFO2(ins), 0, TAG_INLINE },
	{ TAGINFO2(del), 0, TAG_INLINE },

	/* ==== lists ==== */

	{ TAGINFO2(dl), 0, TAG_BLOCK },
	{ TAGINFO2(dt), 2, TAG_BLOCK },
	{ TAGINFO2(dd), 2, TAG_BLOCK },
	{ TAGINFO1(ol), 0, TAG_BLOCK },
	{ TAGINFO1(ul), 0, TAG_BLOCK },
	{ TAGINFO1(li), 2, TAG_BLOCK },
	
	/* ==== form ==== */
	{ TAGINFO2(form), 0, TAG_BLOCK },
	{ TAGINFO2(label), 0, TAG_INLINE },
	{ TAGINFO2(input), 2, TAG_INLINE },
	{ TAGINFO2(select), 0, TAG_INLINE },
	{ TAGINFO2(optgroup), 2, TAG_NODISP },
	{ TAGINFO2(option), 2, TAG_NODISP },
	{ TAGINFO2(textarea), 0, TAG_INLINE },
	{ TAGINFO2(fieldset), 0, TAG_INLINE },
	{ TAGINFO2(legend), 0, TAG_INLINE },
	{ TAGINFO2(button), 0, TAG_INLINE },

	/* ==== table ==== */
	{ TAGINFO2(table), 0, TAG_NODISP },
	{ TAGINFO2(caption), 0, TAG_INLINE },
	{ TAGINFO2(thead), 2, TAG_BLOCK },
	{ TAGINFO2(tfoot), 2, TAG_BLOCK },
	{ TAGINFO2(tbody), 3, TAG_BLOCK },
	{ TAGINFO2(colgroup), 2, TAG_BLOCK },
	{ TAGINFO2(col), 2, TAG_BLOCK },
	{ TAGINFO2(tr), 2, TAG_BLOCK },
	{ TAGINFO2(th), 2, TAG_INLINE },
	{ TAGINFO2(td), 2, TAG_INLINE },
	
	/* ==== document head ==== */
	{ TAGINFO2(head), 3, TAG_NODISP },
	{ TAGINFO2(title), 3, TAG_NODISP },
	{ TAGINFO2(base), 2, TAG_NODISP },
	{ TAGINFO2(meta), 2, TAG_NODISP },
	{ TAGINFO2(style), 0, TAG_NODISP },
	{ TAGINFO2(script), 0, TAG_NODISP },
	{ TAGINFO2(noscript), 0, TAG_NODISP },

	/* document structure */
	{ TAGINFO2(html), 3, TAG_NODISP },


	{ TAGINFO2(_start_), 0, TAG_BLOCK },

	{ NULL, NULL, NULL, 0, NULL, 0, TAG_NODISP }
};

static u_int
hash(char *s)
{
	u_int hashval;
	for (hashval = 0; *s != '\0'; s++)
        hashval = *s + 31*hashval;
    return hashval % HASHSIZE;
}

TagInfo*
GetTagInfo(char *s)
{
    TagInfo *p;
    for (p = hashtab[hash(s)]; p != NULL; p = p->next)
        if (strcmp(s, p->name) == 0)
            return p;
    return NULL;
}

void
InitTagInfo(void)
{
    TagInfo *p;
    u_int hashval;
	/* entry taginfo */
    for (p = taginfo_table; p->name != NULL; p++) {
		hashval = hash(p->name);
		p->next = hashtab[hashval];
		hashtab[hashval] = p;
    }
} 
