#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "XMaidP.h"

#define WIDTH 		(xmw->core.width)
#define HEIGHT		(xmw->core.height)
#define MENUX		(xmw->xmaid.menu_x)
#define MENUY		(xmw->xmaid.menu_y)
#define MENUDX		(xmw->xmaid.menu_dx)
#define MENUDY		(xmw->xmaid.menu_dy)

#define PATNUM 4

typedef struct DefaultData {
	char *names[PATNUM];
	Animate *data[PATNUM];
	Boolean alwaystop;
} DefaultData;

static XrmOptionDescRec options[] = {
	{"-alwaystop", 	".default.top", XrmoptionNoArg, (XtPointer)"True"},
	{"-at", 		".default.top", XrmoptionNoArg, (XtPointer)"True"},
};

#define offset(field)	XtOffsetOf(DefaultData, field)
static XtResource resources[] = {
	{ 
		"top", "Top",
		XtRBoolean, sizeof(Boolean), 
		offset(alwaystop),
		XtRImmediate, (XtPointer)FALSE
	},
	{
		"forepat", "Pattern",
		XtRString, sizeof(String),
		offset(names[0]),
		XtRImmediate, (XtPointer)"maid_f.gif"
	},
	{
		"backpat", "Pattern",
		XtRString, sizeof(String),
		offset(names[1]),
		XtRImmediate, (XtPointer)"maid_b.gif"
	},
	{
		"rightpat", "Pattern",
		XtRString, sizeof(String),
		offset(names[2]),
		XtRImmediate, (XtPointer)"maid_r.gif"
	},
	{
		"leftpat", "Pattern",
		XtRString, sizeof(String),
		offset(names[3]),
		XtRImmediate, (XtPointer)"maid_l.gif"
	},
};
#undef offset

#define ANIM (moddat->anim)
#define DATA (moddat->data)
#define XPOS (moddat->x)
#define YPOS (moddat->y)

static void
init(XMaidWidget xmw, XMaidModuleData *moddat)
{
	DefaultData *p = DATA;

 	int i;
	FILE    *fp;
	Animate *a;

	for (i=0;i<PATNUM;i++) {
		p->data[i] = NULL;
		if ((fp = fopen(XMaidDataPath(xmw,p->names[i]), "r"))) {
			a = animate_new();
			if (!anim_gif_read_stream(a,fp)) {
				p->data[i] = a;
			} else {
				fprintf(stderr, "cant read animated gif: %s\n", p->names[i]);
				exit(1);
			}
			fclose(fp);
		} else {
			perror(p->names[i]);
			exit(1);
		}
	}
	ANIM = p->data[0];

	MENUX =  ANIM->w /2;
	MENUY =  ANIM->h /2;
	MENUDX = ANIM->w /2;
	MENUDY = ANIM->h /2;
}

static void
destroy(XMaidWidget xmw, XMaidModuleData *moddat)
{
	DefaultData *p = DATA;
	int i;
	for (i=0;i<PATNUM;i++)
		if (p->data[i])
			animate_delete(p->data[i]);
}

static int
move(XMaidWidget xmw, XMaidModuleData *moddat)
{
	DefaultData *p = DATA;
	
	Window r_ret, c_ret;
	int p_rx_ret, p_ry_ret, p_x_ret, p_y_ret;
	u_int mask_ret;
	int x, y;
	int distance;
	int p_dx, p_dy;
	int dx, dy;

	XQueryPointer(XtDisplay(xmw), XtWindow(xmw), &r_ret, &c_ret, 
				  &p_rx_ret, &p_ry_ret, &p_x_ret, &p_y_ret, &mask_ret);
	x = XPOS + WIDTH / 2;
	y = YPOS + HEIGHT/ 2;
	p_dx = p_rx_ret - x;
	p_dy = p_ry_ret - y;
	distance = sqrt(p_dx*p_dx + p_dy*p_dy);
	if (distance > 16) {
		dx = 8 * p_dx / distance;
		dy = 8 * p_dy / distance;

		XPOS += dx;
		YPOS += dy;
		
		if (abs(p_dy) > abs(p_dx)) {
			if (p_dy >= 0)
				ANIM = p->data[0];
			else
				ANIM = p->data[1];
		} else {
			if (p_dx >= 0)
				ANIM = p->data[2];
			else
				ANIM = p->data[3];
		}
	} else
		ANIM = p->data[0];

	if (p->alwaystop)
		XRaiseWindow(XtDisplay(xmw), XtWindow(xmw));

	return 0;
}

XMaidModuleInfo default_info = {
	"default",					/* name 			   */
	sizeof(DefaultData),		/* data size           */
	NULL,						/* class initialize	   */
	init,						/* initialize          */
	NULL,						/* duplicate 		   */
	destroy,					/* destroy             */
	move,						/* move                */
	NULL,						/* after_move		   */
	NULL,						/* event handler       */
	None,						/* event mask          */
	TRUE,						/* override redirect   */
	TRUE,						/* mapped when managed */
	options,					/* options             */
	XtNumber(options),			/* options number      */
	resources,					/* resources           */
	XtNumber(resources),		/* resource numvber    */
	1							/* update_timing 	   */
};

static DefaultData data;

XMaidModuleData xmaid_default_moduledata = {
	&data,
	NULL,
	0, 0,
	0, 0,
	None, None
};

XMaidModule xmaid_default_module = {
	NULL, 
	&default_info
};
