/*
 *		XMulti Sample Module
 * 		OSUWARI MULTI
 */

#include <stdio.h>
#include <stdlib.h>

#include "XMaidP.h"
#include <X11/Xmu/Converters.h>
#include "xmaidpop.h"
#include "xmaidmod.h"

#define WIDTH 		(xmw->core.width)
#define HEIGHT		(xmw->core.height)
#define MENUX		(xmw->xmaid.menu_x)
#define MENUY		(xmw->xmaid.menu_y)
#define MENUDX		(xmw->xmaid.menu_dx)
#define MENUDY		(xmw->xmaid.menu_dy)

typedef struct FocusWinData {
	Window oldfocus;
	int oldx, oldy;
	XtJustify justify;
	int yoff;
	String modulename;
	XMaidModule 	*module;
	XMaidModuleData *moduledata;
	char *name;
} FocusWinData;

static XrmOptionDescRec options[] = {
	{"-focusmodule",	".focuswin.focusmodulename", XrmoptionSepArg, NULL},
	{"-justify",		".focuswin.justify", 		 XrmoptionSepArg, NULL},
	{"-pattern",		".focuswin.pattern",		 XrmoptionSepArg, NULL},
	{"-yoff",			".focuswin.yoff", 			 XrmoptionSepArg, NULL},
};

#define offset(field)	XtOffsetOf(FocusWinData, field)
static XtResource resources[] = {
	{
		XtNjustify, XtCJustify,
		XtRJustify, sizeof(XtJustify),
		offset(justify), XtRImmediate, (XtPointer)XtJustifyRight
	},	
	{
		"yoff", "Yoff",
		XtRInt, sizeof(int),
		offset(yoff), XtRImmediate, (XtPointer)-1
	},
	{
		"focusmodulename", XtCModuleName,
		XtRString, sizeof(String),
		offset(modulename), XtRImmediate, (XtPointer)NULL
	},
	{
		"pattern", "Pattern",
		XtRString, sizeof(String),
		offset(name),
		XtRImmediate, (XtPointer)NULL,
	},

};
#undef offset

#define ANIM (moddat->anim)
#define DATA (moddat->data)
#define XPOS (moddat->x)
#define YPOS (moddat->y)

static void
classinit(void)
{
    XtAddConverter( XtRString, XtRJustify, XmuCvtStringToJustify, 
				   (XtConvertArgList)NULL, 0 );
}

static void
init(XMaidWidget xmw, XMaidModuleData *moddat)
{
	FocusWinData *p = DATA;

	p->module   = NULL;
	p->oldfocus = None;
	p->oldx     = 0;
    p->oldy     = 0;

	if (p->modulename) {
		if ((p->module = xmaid_module_open(xmw, p->modulename)))
			p->moduledata = xmaid_moduledata_open(xmw, p->module, 1);
	}

	if (p->module == NULL || p->module->info == NULL) {
		if (p->name) {
			FILE *fp;
			Animate *a;
			if ((fp = XMaidFileOpen(xmw, p->name, "r"))) {
				a = animate_new();
				if (anim_gif_read_stream(a,fp)) {
					animate_delete(a);
					fprintf(stderr, "mod_focusin: cant read  pattern:%s\n", 
							p->name);
					exit(1);
				}
				ANIM   = a;
				MENUX  = ANIM->w/2;
				MENUDX = ANIM->w/4;
				MENUY  = ANIM->h/3;
				MENUDY = 0;	
			} else {
				perror(p->name);
				exit(1);
			}
		} else {
			fprintf(stderr,"mod_focusin: no module!!\n");
			exit(1);
		}

	} else {
		xmaid_module_init(xmw, p->module);
		xmaid_moduledata_init(xmw, p->module, p->moduledata);
		ANIM = p->moduledata->anim;
	}

	if (p->yoff == -1)
		p->yoff = ANIM->h * 2.0 / 3;


}

/* ǡ */
static void
destroy(XMaidWidget xmw, XMaidModuleData *moddat)
{
	FocusWinData *p = DATA;
	if (p->module) {
		xmaid_moduledata_destroy(xmw, p->module, p->moduledata);
		xmaid_module_destroy(xmw, p->module);
	}
}

static int (*old_handler)(Display *dpy, XErrorEvent *ev);

static int
err_handler(Display *dpy, XErrorEvent *ev)
{
	if (ev->error_code == BadWindow)
		return 0;
	else	
		return old_handler(dpy,ev);
}

/* ܥ⥸塼 */
static int
move(XMaidWidget xmw, XMaidModuleData *moddat)
{
	FocusWinData *p = DATA;
	Window focus, work;
	int revert;
	int x, y, width;
	int ret = 200;

	XGetInputFocus(XtDisplay((Widget)xmw), &focus, &revert);

	if (focus == None || focus == PointerRoot)
		focus = p->oldfocus;

	if (focus != None)  {

		Display *dpy = XtDisplay((Widget)xmw);
		XWindowChanges chg;
		Window root, parent;
		Window *children;
		int childnum;

		XSync(dpy,TRUE);
		old_handler = XSetErrorHandler(err_handler);
		for (work = focus,x = 0, y = 0;;) {
			u_int hei, bor, dep;
			if (!XQueryTree(dpy, work, &root, &parent, &children, &childnum)) {
				focus = None;
				goto end;
			}
			XFree(children);
			if (!XGetGeometry(dpy, work, &root, 
							  &x, &y, &width, &hei, &bor, &dep)) {
				focus = None;
				goto end;
			}
			if (parent == root)
				break;
			work = parent;
		}
		XSync(dpy,TRUE);
		XSetErrorHandler(NULL);

		switch (p->justify) {
		case XtJustifyLeft:
			break;
		case XtJustifyCenter:
			x += (width / 2 - WIDTH / 2);
			break;
		case XtJustifyRight:
			x += (width - WIDTH);
			break;
		}
		y -= p->yoff;

		chg.stack_mode = Above;
		chg.sibling    = work;
		XConfigureWindow(dpy, XtWindow(xmw), 
						 CWStackMode | CWSibling, &chg);

		if (p->oldx != x || p->oldy != y) {
			XPOS = x;
			YPOS = y;
		}

	end:
		if (p->oldfocus == None) 
			XtMapWidget((Widget)xmw);

		p->oldfocus = focus;
		p->oldx     = x;
		p->oldy     = y;

	} else {	/* no focus win before */
		XtUnmapWidget((Widget)xmw);
 		p->oldfocus = None;
	}

	if (p->module && p->module->info->move) {
		ret = p->module->info->move(xmw, p->moduledata);
		ANIM = p->moduledata->anim;
	}

	return ret;
}

XMaidModuleInfo focuswin_info = {
	"focuswin",					/* name				   */
	sizeof(FocusWinData),		/* data size           */
	classinit,					/* class initialize	   */
	init,						/* initialize          */
	NULL,						/* duplicate 		   */
	destroy,					/* destroy             */
	move,						/* move                */
	NULL,						/* after_move		   */
	NULL,						/* event handler       */
	None,						/* event mask          */
	TRUE,						/* override redirect   */
	FALSE,						/* mapped when managed */
	options,					/* options             */
	XtNumber(options),			/* options number      */
	resources,					/* resources           */
	XtNumber(resources),		/* resource numvber    */
	0							/* update_timing 	   */
};
