/*
 * 		XMulti Sample Module
 *		OSOUJI MULTI
 */

#include <stdio.h>
#include <stdlib.h>

#include "XMaidP.h"
#include "xmaidpop.h"

#define WIDTH 		(xmw->core.width)
#define HEIGHT		(xmw->core.height)
#define MENUX		(xmw->xmaid.menu_x)
#define MENUY		(xmw->xmaid.menu_y)
#define MENUDX		(xmw->xmaid.menu_dx)
#define MENUDY		(xmw->xmaid.menu_dy)

#define PATNUM 2

typedef struct OsoujiData {
	char *names[PATNUM];
	Animate *data[PATNUM];
	int mopx[PATNUM];
	int mopy;
	int moph;
	char *rootimgname;
	ImageData *rootimg;
	Boolean alwaystop;
	Boolean onroot;
	Boolean drawroot;
	int direction;
} OsoujiData;

static XrmOptionDescRec options[] = {
	{"-alwaystop", 	".osouji.top", 		 XrmoptionNoArg, (XtPointer)"True"},
	{"-at", 		".osouji.top", 		 XrmoptionNoArg, (XtPointer)"True"},
	{"-drawroot", 	".osouji.drawroot",  XrmoptionNoArg, (XtPointer)"True"},
	{"-root", 		".osouji.root", 	 XrmoptionNoArg, (XtPointer)"True"},
	{"-rootimage", 	".osouji.rootimage", XrmoptionSepArg, NULL},
};

#define offset(field)	XtOffsetOf(OsoujiData, field)
static XtResource resources[] = {
	{ 
		"top", "Top",
		XtRBoolean, sizeof(Boolean), 
		offset(alwaystop),
		XtRImmediate, (XtPointer)FALSE
	},
	{ 
		"root", "Root",
		XtRBoolean, sizeof(Boolean), 
		offset(onroot),
		XtRImmediate, (XtPointer)FALSE
	},
	{ 
		"drawroot", "DrawRoot",
		XtRBoolean, sizeof(Boolean), 
		offset(drawroot),
		XtRImmediate, (XtPointer)FALSE
	},
	{ 
		"rootimage", "Rootimage",
		XtRString, sizeof(String), 
		offset(rootimgname),
		XtRImmediate, (XtPointer)"xmultimark.gif"
	},
	{
		"leftpat", "Pattern",
		XtRString, sizeof(String),
		offset(names[0]),
		XtRImmediate, (XtPointer)"souji_l.gif"
	},
	{
		"rightpat", "Pattern",
		XtRString, sizeof(String),
		offset(names[1]),
		XtRImmediate, (XtPointer)"souji_r.gif"
	},
	{
		"lmopX", "MopX",
		XtRInt, sizeof(int),
		offset(mopx[0]),
		XtRImmediate, (XtPointer)16
	},
	{
		"rmopX", "MopX",
		XtRInt, sizeof(int),
		offset(mopx[1]),
		XtRImmediate, (XtPointer)(128-16-20)
	},
	{
		"mopY", "MopY",
		XtRInt, sizeof(int),
		offset(mopy),
		XtRImmediate, (XtPointer)108
	},
	{
		"mopH", "MopH",
		XtRInt, sizeof(int),
		offset(moph),
		XtRImmediate, (XtPointer)16
	},
};
#undef offset

#define ANIM (moddat->anim)
#define DATA (moddat->data)
#define XPOS (moddat->x)
#define YPOS (moddat->y)

static void
init(XMaidWidget xmw, XMaidModuleData *moddat)
{
	OsoujiData *p = DATA;

 	int i;
	FILE    *fp;
	Animate *a;

	for (i=0;i<PATNUM;i++) {
		p->data[i] = NULL;
		if ((fp = XMaidFileOpen(xmw, p->names[i], "r"))) {
			a = animate_new();
			if (!anim_gif_read_stream(a,fp)) {
				p->data[i] = a;
			} else
				animate_delete(a);
			fclose(fp);
		} else {
			perror(p->names[i]);
			exit(1);
		}
	}
	p->direction = 1;
	ANIM = p->data[p->direction];

	p->rootimg = NULL;
	if (p->onroot && p->drawroot) {
		if ((fp = XMaidFileOpen(xmw, p->rootimgname, "r"))) {
			p->rootimg = image_new();
			if (image_read_stream(p->rootimg, fp)) {
				image_delete(p->rootimg);
				p->rootimg = NULL;
			} else {
				image_set_col(p->rootimg, XtDisplay((Widget)xmw),
							  RootWindowOfScreen(XtScreen((Widget)xmw)));
				image_pixmap(p->rootimg, NULL);
				image_free_data(p->rootimg);
			}
		} else {
			perror(p->rootimgname);
			exit(1);
		}
	}

	MENUX =  ANIM->w /2;
	MENUY =  ANIM->h /2;
	MENUDX = ANIM->w /2;
	MENUDY = ANIM->h /2;
}

static void
dup(XMaidWidget xmw, XMaidModuleData *src, XMaidModuleData *moddat)
{
	OsoujiData *s = src->data;
	OsoujiData *p = DATA;

 	int i;
	*p = *s;

	for (i=0;i<PATNUM;i++)
		p->data[i] = animate_dup(s->data[i]);

	p->direction = random() % 2;
	ANIM = p->data[p->direction];
	YPOS = (random() % HeightOfScreen(XtScreen((Widget)xmw))) - ANIM->h;
	XPOS = (random() % (WidthOfScreen(XtScreen((Widget)xmw)) - ANIM->w));
}

static void
destroy(XMaidWidget xmw, XMaidModuleData *moddat)
{
	OsoujiData *p = DATA;

	int i;
	for (i=0;i<PATNUM;i++)
		if (p->data[i])
			animate_delete(p->data[i]);
	if (p->rootimg)
		image_delete(p->rootimg);
}

static int
move(XMaidWidget xmw, XMaidModuleData *moddat)
{
	OsoujiData *p = DATA;
	
	int width = WidthOfScreen(XtScreen((Widget)xmw));
	int height = HeightOfScreen(XtScreen((Widget)xmw));
	int x = XPOS;
	int y = YPOS;

	if (p->direction == 0) {
		x -= 8;
		if (x < -WIDTH-8) {
			p->direction = 1;
			y = (random() % height) - HEIGHT;
			ANIM = p->data[1];
		}
	} else {
		x += 8;
		if (x > width+8) {
			p->direction = 0;
			y = (random() % height) - HEIGHT;
			ANIM = p->data[0];
		}
	}

	if (!p->onroot && p->alwaystop)
		XRaiseWindow(XtDisplay(xmw), XtWindow(xmw));

	XPOS = x;
	YPOS = y;
	return 0;
}

static void
after_move(XMaidWidget xmw, XMaidModuleData *moddat)
{
	OsoujiData *p = DATA;

	if (p->onroot && p->rootimg) {
		int dx, dy;
		int posx = p->mopx[p->direction];
		int posy = p->mopy;
		int x = XPOS + posx;
		int y = YPOS + posy;

		while (x < 0)
			x += p->rootimg->width;
		while (y < 0)
			y += p->rootimg->height;
		x %= p->rootimg->width;
		y %= p->rootimg->height;

		dx = p->rootimg->width - x;
		dy = p->rootimg->height - y;

		if (dx > 20) dx = 20;
		if (dy > p->moph) dy = p->moph;
		
		XCopyArea(XtDisplay(xmw), p->rootimg->pixmap, 
				  moddat->root_work, xmw->xmaid.root_gc,
				  x, y, dx, dy, posx, posy);
			
		if (dx < 20)
			XCopyArea(XtDisplay(xmw), p->rootimg->pixmap, 
					  moddat->root_work, xmw->xmaid.root_gc,
					  0, y, 20 - dx, dy,
					  posx + dx, posy);

		if (dy < p->moph)
			XCopyArea(XtDisplay(xmw), p->rootimg->pixmap,
					  moddat->root_work, xmw->xmaid.root_gc,
					  x, 0, dx, p->moph - dy,
					  posx, posy + dy);
		
		if (dx < 20 && dy < p->moph)
			XCopyArea(XtDisplay(xmw), p->rootimg->pixmap,
					  moddat->root_work, xmw->xmaid.root_gc,
					  0, 0, 20 - dx, p->moph - dy,
					  posx + dy, posy + dy);
	}
}

XMaidModuleInfo osouji_info = {
	"osouji",					/* name                */
	sizeof(OsoujiData),			/* data size           */
	NULL,						/* class initialize	   */
	init,						/* initialize          */
	dup,						/* duplicate 		   */
	destroy,					/* destroy             */
	move,						/* move                */
	after_move,					/* after_move		   */
	NULL,						/* event handler       */
	None,						/* event mask          */
	TRUE,						/* override redirect   */
	TRUE,						/* mapped when managed */
	options,					/* options             */
	XtNumber(options),			/* options number      */
	resources,					/* resources           */
	XtNumber(resources),		/* resource numvber    */
	1							/* update_timing 	   */
};

